from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import bz2
from collections import namedtuple
import timeit

try:
    from cPickle import dumps
except ImportError:
    from pickle import dumps

from dask.core import istask, ishashable

TaskAnnotation = namedtuple("TaskAnnotation", ["an"])


class Task(tuple):
    __slots__ = ()

    _index_names = ("function", "args", "kwargs", "annotations")

    def __new__(cls, fn, *args, **kwargs):
        annots = kwargs.pop("annotations", None)
        return tuple.__new__(Task, (fn, args, kwargs, annots))

    @property
    def function(self):
        return self[0]

    @property
    def args(self):
        return self[1]

    @property
    def kwargs(self):
        return self[2]

    @property
    def annotations(self):
        return self[3]

    def __repr__(self):
        details = ", ".join("%s=%s" % (n, repr(self[i]))
                            for i, n in enumerate(self._index_names)
                            if self[i])
        return 'Task({})'.format(details)


def execute_task_std(arg, cache, dsk=None):
    if isinstance(arg, list):
        return [execute_task_std(a, cache) for a in arg]
    elif istask(arg):
        func, args = arg[0], arg[1:]
        args2 = [execute_task_std(a, cache) for a in args]
        return func(*args2)
    elif not ishashable(arg):
        return arg
    elif arg in cache:
        return cache[arg]
    else:
        return arg


def execute_task_annots(arg, cache, dsk=None):
    if isinstance(arg, list):
        return [execute_task_annots(a, cache) for a in arg]
    elif istask(arg):
        # Strip out annotations for the moment
        # https://github.com/dask/dask/issues/3783
        if type(arg[-1]) == TaskAnnotation:
            func, args, annots = arg[0], arg[1:-1], arg[-1]
        else:
            func, args, annots = arg[0], arg[1:], None
        args2 = [execute_task_annots(a, cache) for a in args]
        return func(*args2)
    elif not ishashable(arg):
        return arg
    elif arg in cache:
        return cache[arg]
    else:
        return arg


def execute_task_obj(arg, cache, dsk=None):
    if isinstance(arg, list):
        return [execute_task_obj(a, cache) for a in arg]
    elif type(arg) == Task:
        # function, args + kwargs properties
        args2 = [execute_task_obj(a, cache) for a in arg[1]]
        kwargs2 = {k: execute_task_obj(v) for k, v in arg[2]}
        return arg[0](*args2, **kwargs2)
    elif istask(arg):
        func, args = arg[0], arg[1:]
        args2 = [execute_task_obj(a, cache) for a in args]
        return func(*args2)
    elif not ishashable(arg):
        return arg
    elif arg in cache:
        return cache[arg]
    else:
        return arg


def f(x, *args, **kwargs):
    return x + 1


units = {"nanoseconds": 1e-9,
         "microseconds": 1e-6,
         "milliseconds": 1e-3,
         "seconds": 1.0}


def format_time(t):
    precision = 3
    scales = [(scale, unit) for unit, scale in units.items()]
    scales.sort(reverse=True)

    for scale, unit in scales:
        if t >= scale:
            break

    return "%.*g %s" % (precision, t / scale, unit)


def create_tasks(N, style="task"):
    if style.lower() == "task":
        return {('task', i): Task(f, i, bob="foo",
                                  annotations={"resource": {"GPU": 1},
                                               "size": i})
                for i in range(N)}
    elif style.lower() == "annotated_tuple":
        return {('task', i): (f, i, {"bob": "foo"},
                              TaskAnnotation({"resource": {"GPU": 1},
                                              "size": i}))
                for i in range(N)}
    elif style.lower() == "tuple":
        return {('task', i): (f, i, {"bob": "foo"},
                              {"resource": {"GPU": 1}, "size": i})
                for i in range(N)}
    else:
        raise ValueError("Invalid style %s" % style)


def create_parser():
    p = argparse.ArgumentParser()
    p.add_argument("-l", "--loops", default=1000, type=int)
    p.add_argument("-t", "--tasks", default=1000, type=int)
    return p


if __name__ == "__main__":
    args = create_parser().parse_args()
    ntasks = args.tasks
    nloops = args.loops
    styles = ("tuple", "annotated_tuple", "task")

    print("\nTASK EXECUTION TIMING")

    exec_tasks_fns = ('execute_task_std',
                      'execute_task_annots',
                      'execute_task_obj')

    for style, exec_fn in zip(styles, exec_tasks_fns):
        # Patch dask.local._execute_task and create_Tasks
        setup = ("from __main__ import create_tasks, TaskAnnotation; "
                 "from __main__ import {exec_fn}; "
                 "import dask; "
                 "from mock import patch; "
                 "task_patch = patch('dask.local._execute_task', "
                 "                   new={exec_fn}); "
                 "task_patch.start(); "
                 "tasks = create_tasks({ntasks}, '{style}') "
                 .format(exec_fn=exec_fn, ntasks=ntasks, style=style))

        t = timeit.timeit("dask.get(tasks, tasks.keys())",
                          setup=setup, number=nloops)
        print(style, format_time(t / (nloops*ntasks)))

    print("\nTASK CREATION TIMING")

    for style in styles:
        setup = "from __main__ import create_tasks"
        t = timeit.timeit("create_tasks(%d, '%s')" % (ntasks, style),
                          setup=setup, number=nloops)

        print(style, format_time(t / (nloops*ntasks)))

    print("\nPICKLING TASK TIMING")

    for style in styles:
        setup = ("from __main__ import create_tasks, TaskAnnotation; "
                 "from cPickle import dumps; "
                 "tasks = create_tasks(%d, '%s') " % (ntasks, style))

        t = timeit.timeit("dumps(tasks)", setup=setup, number=nloops)

        print(style, format_time(t / (nloops*ntasks)))

    print("\nPICKLING SIZES")

    for style in styles:
        print(style, len(dumps(create_tasks(ntasks, style))), "bytes")

    print("\nBZIP2 PICKLED SIZES")

    for style in styles:
        nbytes = len(bz2.compress(dumps(create_tasks(ntasks, style))))
        print(style, nbytes, "bytes")
